from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from pycdp.core.connection import Connection
from pycdp.core.session import TargetSession
from pycdp.core.events import EventStream

if TYPE_CHECKING:
    from pycdp.core.chrome import Chrome


@dataclass
class TargetInfo:
    """
    Represents information about a target.
    """
    id: str
    title: str
    description: str
    url: str
    type: str
    web_socket_debugger_url: str
    parent_id: str = None
    favicon_url: str = None


@dataclass
class Target:
    """
    Represents a CDP target.
    """
    _connection: Connection = field(
        init=False,
        repr=False
    )

    chrome: 'Chrome'
    info: TargetInfo

    @property
    def is_connected(self):
        """
        Public readonly access to the connection status.
        """
        return self._connection.is_connected

    @property
    def ws_url(self):
        """
        A readonly property Web socket url.
        """
        return f'ws://{self.chrome.host}:{self.chrome.port}/devtools/page/{self.info.id}'

    def __post_init__(self):
        """
        Assigns the connection to the target.
        """
        self._connection = Connection(self.ws_url)

    async def close_session(self, session: 'TargetSession'):
        """
        Closes the session. Calls `TargetSession.close`.
        """
        await session.close()

    def close_stream(self, stream: EventStream):
        """
        Closes the stream. Calls `Connection.close_stream`.
        """
        return self._connection.close_stream(stream)

    async def connect(self):
        """
        Connects to the target. Calls `Connection.connect`.
        """
        return await self._connection.connect()

    async def disconnect(self):
        """
        Disconnects from the target. Calls `Connection.disconnect`.
        """
        return await self._connection.disconnect()

    def open_stream(self, events: list[str]):
        """
        Opens a stream. Calls `Connection.open_stream`.
        """
        return self._connection.open_stream(events)

    async def open_session(self):
        """
        Opens a session with the target.
        """
        session = TargetSession.create(self)
        await session.open()

        return session

    async def send(self, method: str, params: dict = None):
        """
        Sends a message to the target. Calls `Connection.send`.
        """
        return await self._connection.send(
            method,
            params or {}
        )

    async def send_and_await_response(self, method: str, params: dict = None):
        """
        Sends a message to the target and awaits a response. Calls `Connection.send_and_await_response`
        """
        return await (await self.send(
            method,
            params
        ))



from typing import Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from pycdp import Target


class PYCDPException(Exception):
    """
    Base exception class for all pycdp exceptions.
    """


class InvalidRPCResponse(PYCDPException):
    """
    Exception raised when an invalid rpc response is received.
    """


class NoTargetFound(PYCDPException):
    """
    Exception raised when no target is found.
    """


class NoTargetFoundMatchingCondition(NoTargetFound):
    """
    Exception raised when no target is found matching the given condition.
    """


def raise_invalid_rpc_response(rpc_response: dict):
    """
    Raises an InvalidRPCResponse exception.
    """
    raise InvalidRPCResponse(
        f'Invalid rpc response. '
        f'Response: {rpc_response}'
    )


def raise_no_target_found():
    """
    Raises a NoTargetFound exception.
    """
    raise NoTargetFound(
        f'No target found.'
    )


def raise_no_target_found_matching_condition(condition: Callable[['Target'], bool]):
    """
    Raises a NoTargetFoundMatchingCondition exception.
    """
    raise NoTargetFoundMatchingCondition(
        f'No target found matching the given condition. '
        f'Condition: {condition.__name__}'
    )



from .core import (
    Chrome,
    Target,
    TargetSession,
    Connection,
    EventStream,
    EventStreamReader
)
from .exceptions import (
    PYCDPException,
    InvalidRPCResponse,
    NoTargetFound,
    NoTargetFoundMatchingCondition
)



from typing import Literal

LoggableEvents = Literal[
    '*',
    'connection.handle_message',
    'connection.handle_event',
    'connection.handle_response',
    'connection.send'
]

events_to_log: dict[LoggableEvents, bool] = {
    '*': False,
    'connection.handle_message': False,
    'connection.handle_event': False,
    'connection.handle_response': False,
    'connection.send': False
}


def enable_logging(
        events: list[LoggableEvents] = None
):
    events = events or []

    for event in events:
        events_to_log[event] = True


def disable_logging(
        events: list[LoggableEvents] = None
):
    events = events or []

    for event in events:
        events_to_log[event] = False


def is_logging_enabled(
        event: LoggableEvents
):
    return events_to_log['*'] or events_to_log[event]



import asyncio
from dataclasses import dataclass, field
from typing import TypeVar, AsyncGenerator, TYPE_CHECKING
from uuid import uuid4

if TYPE_CHECKING:
    from pycdp import Connection

_T = TypeVar('_T')


@dataclass(eq=False)
class EventStream:
    """
    Represents an asynchronous stream of CDP events received from the connection.
    """
    connection: 'Connection'
    events: list[_T]
    events_to_listen: list[str]
    reader: 'EventStreamReader | None'
    next: asyncio.Future

    @classmethod
    def create(cls, connection: 'Connection', events_to_listen: list[str]):
        """
        Creates a new instance of the event stream.
        """
        loop = asyncio.get_event_loop()

        return cls(
            connection=connection,
            events=[],
            events_to_listen=events_to_listen,
            reader=None,
            next=loop.create_future(),
        )

    @property
    def is_closed(self):
        """
        Public readonly access to the closed status of the stream.
        """
        return self.connection.is_stream_closed(self)

    def __enter__(self):
        """
        Allows this object to be used as a context manager.
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Closes this stream when used as a context manager.
        """
        self.close()

        if exc_type is not None:
            return False

    def close(self):
        """
        Unsubscribes this stream from receiving events from the connection.
        """
        self.connection.close_stream(self)

    def get_reader(self):
        """
        Creates the reader for this stream.

        NOTES:
            - This method returns the the same instance for the lifetime of this stream.
        """

        if self.reader is None:
            self.reader = EventStreamReader(self)

        return self.reader

    async def iterate(self) -> AsyncGenerator[_T, None]:
        """
        Iterates over all the events in this stream and asynchronously yields new events as they are received.

        NOTES:
            - Should not be used by public API. First call `EventStream.get_reader`
              and then use `EventStreamReader.iterate` instead.
        """
        for item in self.events:
            yield item

        while True:
            yield await self.next

    def write(self, item: _T):
        """
        Writes an item to the stream. Responsible for resolving futures.
        """
        loop = asyncio.get_event_loop()

        self.events.append(item)
        self.next.set_result(item)
        self.next = loop.create_future()


@dataclass
class EventStreamReader:
    """
    Object that has read only access to an event stream.
    """
    stream: EventStream

    @property
    def events_to_listen(self) -> list[str]:
        """
        Public readonly access to the CDP events to listen for. Provided by the stream.
        """
        return self.stream.events_to_listen

    @property
    def connection(self):
        """
        Public readonly access to the `Connection` instance. Provided by the stream.
        """
        return self.stream.connection

    @property
    def is_closed(self):
        """
        Public readonly access to the closed status of the stream. Provided by the stream.
        """
        return self.stream.is_closed

    def __enter__(self):
        """
        Allows this object to be used as a context manager.
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Closes this stream when used as a context manager.
        """
        self.close()

        if exc_type is not None:
            return False

    def close(self):
        """
        Closes this stream.
        """
        self.stream.close()

    async def iterate(self) -> AsyncGenerator[_T, None]:
        """
        Iterates over the events asynchronously in this stream.
        """
        async for item in self.stream.iterate():
            yield item



import os
from dataclasses import dataclass
from typing import Callable, Any

import requests

from pycdp.core.target import Target, TargetInfo
from pycdp.exceptions import NoTargetFoundMatchingCondition
from pycdp.utils import UNDEFINED


@dataclass
class Chrome:
    """
    Represents a Chrome instance on a given host and port.
    """
    host: str
    port: int

    @classmethod
    def create(cls, host: str = '127.0.0.1', port: int = 9222):
        """
        Method to create a new instance of Chrome with default host and port.
        """
        return cls(
            host=host,
            port=port
        )

    def start(self, cli_args: list[str] = None):
        """
        Starts chrome through the command line.
        """
        commands = [
            f'start chrome',
            f'--remote-debugging-port={self.port}',
            f'--remote-allow-origins=*'
        ]

        commands.extend(cli_args or [])
        command = ' '.join(commands)

        os.system(command)

        return self

    def close_tab(self, target_id: str):
        """
        Closes the tab with the given target id.
        """
        url = f'http://{self.host}:{self.port}/json/close/{target_id}'

        response = requests.get(url)
        response.raise_for_status()

    def get_first_target(
            self,
            condition: Callable[[Target], bool] = None,
            default: Any = UNDEFINED
    ) -> Target:
        """
        Fetches and returns the first target. Raises an exception if no target is found and no default is supplied.

        NOTES:
            Supply a condition to filter against each target.
            Supply a default to return a default if no target is found.
        """
        for target in self.iterate_targets():
            if condition:
                if condition(target):
                    return target
            else:
                return target

        if default is not UNDEFINED:
            return default

        raise NoTargetFoundMatchingCondition(
            condition
        )

    def get_targets(self) -> list[Target]:
        """
        Fetches a list of all the targets.
        """
        return list(
            self.iterate_targets()
        )

    def iterate_targets(self):
        """
        Returns a generator that fetches and iterates over all the targets.
        """
        url = f'http://{self.host}:{self.port}/json/list'

        response = requests.get(url)
        response.raise_for_status()

        for target in response.json():
            yield Target(
                chrome=self,
                info=TargetInfo(
                    id=target['id'],
                    title=target['title'],
                    description=target['description'],
                    url=target['url'],
                    type=target['type'],
                    web_socket_debugger_url=target['webSocketDebuggerUrl']
                )
            )

    def open_tab(self, tab_url: str = None) -> Target:
        """
        Opens a new tab.
        """
        url = f'http://{self.host}:{self.port}/json/new'

        if tab_url:
            url += f'?{tab_url}'

        response = requests.put(url)
        response.raise_for_status()

        return Target(
            chrome=self,
            info=TargetInfo(
                id=response.json()['id'],
                title=response.json()['title'],
                description=response.json()['description'],
                url=response.json()['url'],
                type=response.json()['type'],
                web_socket_debugger_url=response.json()['webSocketDebuggerUrl']
            )
        )



from dataclasses import dataclass
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from pycdp.core.events import EventStream
    from pycdp.core.target import Target


@dataclass
class TargetSession:
    """
    Represents a session with a target.
    """
    target: 'Target'
    session_id: str | None

    @classmethod
    def create(cls, target: 'Target'):
        """
        Creates a new instance of the target session.
        """
        return cls(
            target=target,
            session_id=None
        )

    async def __aenter__(self):
        """
        Allows this object to be used as a context manager.
        """
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        """
        Closes this session when used as a context manager.
        """
        await self.close()

        if exc_type is not None:
            return False

    async def close(self):
        """
        Closes the session by detaching from the target.
        """
        method = 'Target.detachFromTarget'
        params = {
            'targetId': self.session_id
        }

        await self.send_and_await_response(
            method,
            params
        )

    def close_stream(self, stream: 'EventStream'):
        """
        Closes the stream. Calls `Target.close_stream`.
        """
        return self.target.close_stream(stream)

    async def open(self):
        """
        Opens the session by attaching to the target.
        """
        method = 'Target.attachToTarget'
        params = {
            'targetId': self.target.info.id
        }

        result = await self.send_and_await_response(
            method,
            params
        )

        self.session_id = result['sessionId']

    def open_stream(self, events: list[str]):
        """
        Opens a stream. Calls `Target.open_stream`.
        """
        return self.target.open_stream(events)

    async def send(self, method: str, params: dict = None):
        """
        Sends a message to the target. Calls `Target.send`.
        """
        params = params or {}
        params['sessionId'] = self.session_id

        return await self.target.send(
            method,
            params
        )

    async def send_and_await_response(self, method: str, params: dict = None):
        """
        Sends a message to the target and awaits a response. Calls `Target.send_and_await_response`
        """
        params = params or {}
        params['sessionId'] = self.session_id

        return await self.target.send_and_await_response(
            method,
            params
        )



import asyncio
import json
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Optional

import websockets.client as websockets

from pycdp.core.events import EventStreamReader, EventStream
from pycdp import logging
from pycdp.exceptions import raise_invalid_rpc_response

_id = 0


def next_rpc_id():
    """
    Generates a new sequential id for rpc requests.
    """
    global _id
    _id += 1
    return _id


def validate_rpc_response(response: dict):
    """
    Validates the given rpc response and raises an exception if the response contains an error.
    """
    if 'error' in response and response['error']:
        raise_invalid_rpc_response(response)


@dataclass
class Connection:
    """
    Represents a websocket connection to a CDP target.
    """

    """
    The url of the websocket connection.
    """
    ws_url: str

    """
    The futures that are currently waiting for a response to a request.
    """
    in_flight_futures: dict[int, asyncio.Future] = field(
        default_factory=dict,
        init=False
    )

    """
    The streams that are currently listening for events.
    """
    streams: defaultdict[str, list[EventStream]] = field(
        default_factory=lambda: defaultdict(list),
        init=False
    )

    """
    The websocket connection instance.
    """
    ws: Optional[websockets.WebSocketClientProtocol] = field(
        default=None,
        init=False
    )

    """
    The asyncio task that is listening for messages from the websocket in the background.
    """
    ws_listener: Optional[asyncio.Task] = field(
        default=None,
        init=False
    )

    """
    The future that is set when the websocket connection is established.
    """
    ws_connected: Optional[asyncio.Future] = field(
        default=None,
        init=False
    )

    @property
    def is_connected(self):
        """
        Public readonly access to the connection status.
        """
        return self.ws is not None

    def _handle_event(self, event: dict):
        """
        Handles an event received from the websocket.
        """
        if logging.is_logging_enabled('connection.handle_event'):
            print(event)

        for stream in self.streams[event['method']]:
            stream.write(event)

    async def _handle_message(self, message: str):
        """
        Handles any message received from the websocket.
        """
        if logging.is_logging_enabled('connection.handle_message'):
            print(message)

        message = json.loads(message)

        if 'id' in message:
            return self._handle_response(
                message
            )

        else:
            return self._handle_event(
                message
            )

    def _handle_response(self, response: dict):
        """
        Handles a response received from the websocket.
        """
        if logging.is_logging_enabled('connection.handle_response'):
            print(response)

        future = self.in_flight_futures.pop(
            response['id'],
            None
        )

        if future is None:
            return

        try:
            validate_rpc_response(
                response
            )

            future.set_result(
                response['result']
            )

        except Exception as e:
            future.set_exception(e)

    async def _listen_async(self):
        """
        Listens for messages from the websocket.
        """
        async with websockets.connect(self.ws_url) as ws:
            self.ws = ws
            self.ws_connected.set_result(
                True
            )

            while True:
                message = await ws.recv()
                await self._handle_message(
                    message
                )

    def close_stream(
            self,
            stream: EventStream
    ):
        """
        Unsubscribes the given event stream from receiving events from the connection.
        """
        for event in stream.events_to_listen:
            self.streams[event].remove(stream)

    async def connect(self):
        """
        Connects to the websocket and starts the listener task.
        """
        if self.is_connected:
            return

        loop = asyncio.get_event_loop()

        self.ws_connected = loop.create_future()
        self.ws_listener = loop.create_task(self._listen_async())

        await self.ws_connected

    async def disconnect(self):
        """
        Disconnects from the websocket and cancels the listener task.
        """
        if not self.is_connected:
            return

        await self.ws.close()
        self.ws_listener.cancel()

    def is_stream_closed(self, stream: EventStream):
        """
        Checks if the given stream is closed.
        """
        for event in stream.events_to_listen:
            if stream in self.streams[event]:
                return False

        return True

    def open_stream(
            self,
            events: list[str]
    ) -> EventStreamReader:
        """
        Opens a new event stream for the given events.
        """
        stream = EventStream.create(
            self,
            events
        )

        for event in events:
            self.streams[event].append(
                stream
            )

        return stream.get_reader()

    async def send(
            self,
            method: str,
            params: dict
    ):
        """
        Sends a message to the websocket.
        """
        loop = asyncio.get_event_loop()

        request_id = next_rpc_id()
        request = {
            'id': request_id,
            'method': method,
            'params': params
        }

        request = json.dumps(request)
        future = loop.create_future()

        self.in_flight_futures[request_id] = future

        try:
            await self.ws.send(request)

        except Exception as e:
            future.set_exception(e)

        return future



from setuptools import setup, find_packages

setup(
    name='aiocdp',
    version='0.0.1',
    packages=find_packages(
        include=[
            'pycdp',
            'pycdp.*'
        ],
        exclude=[
            'tests',
            'tests.*'
        ]
    ),
    install_requires=[
        'websockets',
        'requests',
    ]
)



from typing import TypeVar

T = TypeVar('T')


class _Undefined:
    """
    Represents an undefined value.
    """
    def __repr__(self):
        return '?'

    def __getattr__(self, item):
        return self

    def __bool__(self):
        return False


UNDEFINED = _Undefined()
MaybeUndefined = T | _Undefined


def is_undefined(value):
    return isinstance(value, _Undefined)


def is_defined(value):
    return not is_undefined(value)
